@@ -4,6 +4,15 @@ Omnata Plugin Runtime.
 Includes data container classes and defines the contract for a plugin.
 """
 from __future__ import annotations
+import sys
+from typing import Union
+if tuple(sys.version_info[:2]) >= (3, 9):
+    # Python 3.9 and above
+    from typing import Annotated  # pylint: disable=ungrouped-imports
+else:
+    # Python 3.8 and below
+    from typing_extensions import Annotated
+
 
 import datetime
 import http
@@ -20,7 +29,7 @@ import jinja2
 import pandas
 import pydantic
 import pydantic.json
-from pydantic import parse_obj_as, root_validator
+from pydantic import Field, parse_obj_as, root_validator
 from dateutil.parser import parse
 from jinja2 import Environment
 from pydantic import BaseModel  # pylint: disable=no-name-in-module
@@ -169,6 +178,7 @@ class SyncRequest(ABC):
         :return: nothing
         """
         logger.info(f"Initiating SyncRequest for sync run {run_id}, run deadline: {run_deadline}")
+        self.deadline_reached:bool = False
         self._run_deadline = run_deadline
         self.plugin_instance = plugin_instance
         self._source_app_name = source_app_name
@@ -240,12 +250,12 @@ class SyncRequest(ABC):
 
     def __apply_results_worker(self, cancellation_token:threading.Event):
         """
-        Designed to be run in a thread, this method polls the results every 10 seconds and sends them back to Snowflake.
+        Designed to be run in a thread, this method polls the results every 20 seconds and sends them back to Snowflake.
         """
         while not cancellation_token.is_set():
             logger.info("apply results worker checking for results")
             self.apply_results_queue()
-            cancellation_token.wait(10)
+            cancellation_token.wait(20)
         logger.info("apply results worker exiting")
     
     def __rate_limit_update_worker(self, cancellation_token:threading.Event):
@@ -273,6 +283,7 @@ class SyncRequest(ABC):
             logger.info("cancel checking worker checking for cancellation")
             if (datetime.datetime.now(datetime.timezone.utc) > self._run_deadline):  # pylint: disable=protected-access
                 # if we've reached the deadline for the run, end it
+                self.deadline_reached = True
                 self.apply_deadline_reached()  # pylint: disable=protected-access
                 return
 
@@ -552,7 +563,7 @@ class OutboundSyncRequest(SyncRequest):
                     )
                     # upload all cached apply results
                     all_dfs = pandas.concat(self._apply_results)
-                    logger.info(f"applying: {all_dfs}")
+                    #logger.info(f"applying: {all_dfs}")
                     self._apply_results_dataframe(all_dfs)
                     self._apply_results.clear()
                 else:
@@ -857,6 +868,12 @@ class InboundSyncRequest(SyncRequest):
             s.stream_name: s for s in streams
         }
         self._apply_results: Dict[str, List[pandas.DataFrame]] = {}
+        # named by convention, see SyncRunProcessor.enqueue
+        self._criteria_deletes_table_name = (
+            f"{self._source_app_name}.{self._results_schema_name}.{self._results_table_name}_CRITERIA_DELETES"
+        )
+        # These are similar to the results, but represent requests to delete records by some criteria
+        self._apply_results_criteria_deletes: Dict[str, List[pandas.DataFrame]] = {}
         self._latest_states: Dict[str, Any] = {}
         self._temp_tables = {}
         self._temp_table_lock = threading.Lock()
@@ -879,25 +896,28 @@ class InboundSyncRequest(SyncRequest):
         logger.info("InboundSyncRequest apply_results_queue ")
         if self._apply_results is not None:
             with self._apply_results_lock:
+                results:List[pandas.DataFrame] = []
+                stream_names:List[str] = []
                 for stream_name, stream_results in self._apply_results.items():
-                    results = [
+                    results.extend([
                         x for x in stream_results if x is not None and len(x) > 0
-                    ]  # remove any None/empty dataframes
-                    if len(results) > 0:
-                        logger.info(
-                            f"Applying {len(results)} batches of queued results"
-                        )
-                        # upload all cached apply results
-                        all_dfs = pandas.concat(results)
-                        logger.info(f"applying: {all_dfs}")
-                        self._apply_results_dataframe(stream_name, all_dfs)
-                        # add the count of this batch to the total for this stream
-                        self._stream_record_counts[
-                            stream_name
-                        ] = self._stream_record_counts[stream_name] + len(all_dfs)
-                    # update the stream state object too
-                    self._apply_latest_states()
-                    self._apply_results[stream_name] = None
+                    ])  # remove any None/empty dataframes
+                    stream_names.append(stream_name)
+                if len(results) > 0:
+                    logger.info(
+                        f"Applying {len(results)} batches of queued results"
+                    )
+                    # upload all cached apply results
+                    all_dfs = pandas.concat(results)
+                    #logger.info(f"applying: {all_dfs}")
+                    self._apply_results_dataframe(stream_names, all_dfs)
+                    # add the count of this batch to the total for this stream
+                    self._stream_record_counts[
+                        stream_name
+                    ] = self._stream_record_counts[stream_name] + len(all_dfs)
+                # update the stream state object too
+                self._apply_latest_states()
+                self._apply_results[stream_name] = None
                 self._apply_results = {}
                 # update the inbound stream record counts, so we can see progress
                 self._plugin_message(
@@ -907,6 +927,22 @@ class InboundSyncRequest(SyncRequest):
                         stream_errors=self._omnata_log_handler.stream_global_errors
                     )
                 )
+        # also take care of uploading delete requests
+        if self._apply_results_criteria_deletes is not None:
+            with self._apply_results_lock:
+                results:List[pandas.DataFrame] = []
+                for stream_name, stream_results in self._apply_results_criteria_deletes.items():
+                    results.extend([
+                        x for x in stream_results if x is not None and len(x) > 0
+                    ])
+                if len(results) > 0:
+                    logger.info(
+                        f"Applying {len(results)} batches of queued criteria deletes"
+                    )
+                    # upload all cached apply results
+                    all_dfs = pandas.concat(results)
+                    #logger.info(f"applying: {all_dfs}")
+                    self._apply_criteria_deletes_dataframe(all_dfs)
 
     def apply_cancellation(self):
         """
@@ -942,11 +978,16 @@ class InboundSyncRequest(SyncRequest):
             message=PluginMessageAbandonedStreams(abandoned_streams=abandoned_streams)
         )
 
-    def enqueue_results(self, stream_name: str, results: List[Dict], new_state: Any):
+    def enqueue_results(self, stream_name: str, results: List[Dict], new_state: Any, is_delete:Union[bool,List[bool]] = False):
         """
-        Adds some results to the queue for applying asynchronously
+        Adds some results to the queue for applying asynchronously.
+        stream_name: str, the name of the stream
+        results: List[Dict], the results to enqueue
+        new_state: Any, the new state which applies to the stream, given the new results
+        is_delete: Union[bool,List[bool]], whether the results are deletes or not
+        is_delete can be a single value, which means all results are the same, or a list of booleans, which means each result is different
+        For records where is_delete is True, you can provide the current record value if it is known, or just the identifier
         """
-        # TODO: maybe also have a mechanism to apply immediately if the queued results are getting too large
         logger.info(f"Enqueueing {len(results)} results for upload")
         if stream_name is None or len(stream_name) == 0:
             raise ValueError("Stream name cannot be empty")
@@ -954,10 +995,72 @@ class InboundSyncRequest(SyncRequest):
             existing_results: List[pandas.DataFrame] = []
             if stream_name in self._apply_results:
                 existing_results = self._apply_results[stream_name]
-            existing_results.append(self._preprocess_results_list(stream_name, results))
+            existing_results.append(self._preprocess_results_list(stream_name, results, is_delete))
             self._apply_results[stream_name] = existing_results
             current_latest = self._latest_states or {}
             self._latest_states = {**current_latest, **{stream_name: new_state}}
+        # if the total size of all the dataframes exceeds 200MB, apply the results immediately
+        # we'll use df.memory_usage(index=True) for this
+        if self.development_mode is False:
+            # note: we want to do it for all values in self._apply_results, not just the new one
+            # so first we need to get the list of lists from the dictionary values and flatten it
+            # then we can sum the memory usage of each dataframe
+            # if the total exceeds 200MB, we apply the results immediately
+            all_df_lists:List[List[pandas.DataFrame]] = list(self._apply_results.values())
+            # flatten
+            all_dfs:List[pandas.DataFrame] = [x for sublist in all_df_lists for x in sublist]
+            combined_length = sum([len(x) for x in all_dfs])
+            # first, don't both if the count is less than 10000, since it's unlikely to be even close
+            if combined_length > 10000:
+                if sum([x.memory_usage(index=True).sum() for x in all_dfs]) > 200000000:
+                    logger.info(f"Applying results queue immediately due to combined dataframe size")
+                    self.apply_results_queue()
+    
+    def delete_by_criteria(self, stream_name: str, criteria: Dict[str, Any]):
+        """
+        Submits some critera (field→value dict) which will cause matching records to be marked as deleted at the end of the run. 
+        This feature was created primarily for array fields that become child streams.
+        The parent record is updated, which means there is a set of new children, but we need to delete the previously sync'd records and we don't know their identifiers.
+
+        The criteria is applied before the new records for the current run are applied. In other words, it will not delete any records from the current run.
+
+        For a record to be deleted, it must match fields with all the criteria supplied. At least one field value must be provided.
+        """
+        if len(criteria) == 0:
+            raise ValueError("At least one field value must be provided for deletion criteria")
+        if stream_name not in self._streams_dict:
+            raise ValueError(
+                f"Cannot delete records for stream {stream_name} as its configuration doesn't exist"
+            )
+        # append the new criteria to the self._criteria_deletes_table_name table
+        # this table has two columns:
+        # STREAM_NAME: string
+        # DELETE_CRITERIA: object
+        with self._apply_results_lock:
+            logger.info(
+                f"Enqueuing {len(criteria)} delete criteria for stream {stream_name} for upload"
+            )
+            existing_results: List[pandas.DataFrame] = []
+            if stream_name in self._apply_results_criteria_deletes:
+                existing_results = self._apply_results_criteria_deletes[stream_name]
+            existing_results.append(pandas.DataFrame([{"STREAM_NAME":stream_name,"DELETE_CRITERIA": criteria}]))
+            self._apply_results_criteria_deletes[stream_name] = existing_results
+        # if the total size of all the dataframes exceeds 200MB, apply the results immediately
+        # we'll use df.memory_usage(index=True) for this
+        if self.development_mode is False:
+            # note: we want to do it for all values in self._apply_results_criteria_deletes, not just the new one
+            # so first we need to get the list of lists from the dictionary values and flatten it
+            # then we can sum the memory usage of each dataframe
+            # if the total exceeds 200MB, we apply the results immediately
+            all_df_lists:List[List[pandas.DataFrame]] = list(self._apply_results_criteria_deletes.values())
+            # flatten
+            all_dfs:List[pandas.DataFrame] = [x for sublist in all_df_lists for x in sublist]
+            combined_length = sum([len(x) for x in all_dfs])
+            # first, don't both if the count is less than 10000, since it's unlikely to be even close
+            if combined_length > 10000:
+                if sum([x.memory_usage(index=True).sum() for x in all_dfs]) > 200000000:
+                    logger.info(f"Applying criteria deletes queue immediately due to combined dataframe size")
+                    self.apply_results_queue() 
 
     def mark_stream_complete(self, stream_name: str):
         """
@@ -1018,14 +1121,15 @@ class InboundSyncRequest(SyncRequest):
             logger.debug(f"Failure to convert inbound data: {str(exception)}")
         return data
 
-    def _preprocess_results_list(self, stream_name: str, results: List[Dict]):
+    def _preprocess_results_list(self, stream_name: str, results: List[Dict],is_delete:Union[bool,List[bool]]):
         """
         Creates a dataframe from the enqueued list, ready to upload.
         The result is a dataframe contain all (and only):
         'APP_IDENTIFIER' string
         'STREAM_NAME' string
         'RETRIEVE_DATE' datetime (UTC)
-        'RECORD_DATA' object
+        'RECORD_DATA' object,
+        'IS_DELETED' boolean
         """
         # for required_column in ['RECORD_DATA']:
         #    if required_column not in results_df.columns:
@@ -1036,6 +1140,9 @@ class InboundSyncRequest(SyncRequest):
             )
         logger.info(f"preprocessing for stream: {self._streams_dict[stream_name]}")
         if len(results) > 0:
+            if isinstance(is_delete, list):
+                if len(results) != len(is_delete):
+                    raise ValueError(f"results and is_delete lists must be the same length")
             # We need to remove any values (included nesting) which are empty dicts. This is to prevent the arrow error: 
             # Cannot write struct type '<field_name>' with no child field to Parquet. Consider adding a dummy child field.
             results = [remove_empty_dict_values(result) for result in results]
@@ -1081,11 +1188,20 @@ class InboundSyncRequest(SyncRequest):
                         )
 
                 results_df["APP_IDENTIFIER"] = results_df["RECORD_DATA"].apply(lambda x: get_nested_value(dict(x),primary_key_field))
-            # we jump the record data to a json string to make uploading to Snowflake less error prone
-            results_df["RECORD_DATA"] = results_df["RECORD_DATA"].apply(json.dumps)
             # the timestamps in Snowflake are TIMESTAMP_LTZ, so we upload in string format to ensure the
             # timezone information is present.
             results_df["RETRIEVE_DATE"] = str(datetime.datetime.now().astimezone())
+            # create the IS_DELETED column from the is_delete list
+            results_df["IS_DELETED"] = is_delete
+            # for each record, if IS_DELETED is true and RECORD_DATA only contains a single key, we assume that's the identifier
+            # in this case, we nullify the RECORD_DATA column to indicate that the delete operation does not contain the full record
+            for index, row in results_df.iterrows():
+                if row["IS_DELETED"] and len(row["RECORD_DATA"]) == 1:
+                    results_df.at[index, "RECORD_DATA"] = None
+            # we dump the record data to a json string to make uploading to Snowflake less error prone, but only if it's not None
+            results_df["RECORD_DATA"] = results_df["RECORD_DATA"].apply(
+                lambda x: json.dumps(x) if x is not None else None
+            )
             results_df["STREAM_NAME"] = stream_name
         else:
             results_df = pandas.DataFrame(
@@ -1095,16 +1211,17 @@ class InboundSyncRequest(SyncRequest):
                     "STREAM_NAME",
                     "RECORD_DATA",
                     "RETRIEVE_DATE",
+                    "IS_DELETED"
                 ],
             )
         # trim out the columns we don't need to return
         return results_df[
             results_df.columns.intersection(
-                ["APP_IDENTIFIER", "STREAM_NAME", "RECORD_DATA", "RETRIEVE_DATE"]
+                ["APP_IDENTIFIER", "STREAM_NAME", "RECORD_DATA", "RETRIEVE_DATE", "IS_DELETED"]
             )
         ]
 
-    def _apply_results_dataframe(self, stream_name: str, results_df: pandas.DataFrame):
+    def _apply_results_dataframe(self, stream_names: List[str], results_df: pandas.DataFrame):
         """
         Applies results for an inbound sync. The results are staged into a temporary
         table in Snowflake, so that we can make an atomic commit at the end.
@@ -1137,7 +1254,8 @@ class InboundSyncRequest(SyncRequest):
                 #                                column_order='index',
                 #                                #create_temp_table=True
                 #                                )
-                self._results_exist[stream_name] = True
+                for stream_name in stream_names:
+                    self._results_exist[stream_name] = True
         else:
             logger.info("Results dataframe is empty, not applying")
 
@@ -1148,6 +1266,34 @@ class InboundSyncRequest(SyncRequest):
         """
         self._plugin_message(PluginMessageStreamState(stream_state=self._latest_states))
 
+    def _apply_criteria_deletes_dataframe(self, results_df: pandas.DataFrame):
+        """
+        Applies results for an inbound sync. The results are staged into a temporary
+        table in Snowflake, so that we can make an atomic commit at the end.
+        """
+        if len(results_df) > 0:
+            with self._snowflake_query_lock:
+                logger.info(
+                    f"Applying {len(results_df)} criteria deletes to {self._criteria_deletes_table_name}"
+                )
+                # try setting parquet engine here, since the engine parameter does not seem to make it through to the write_pandas function
+                success, nchunks, nrows, _ = write_pandas(
+                    conn=self._session._conn._cursor.connection,  # pylint: disable=protected-access
+                    df=results_df,
+                    table_name=self._criteria_deletes_table_name,
+                    quote_identifiers=False,  # already done in get_temp_table_name
+                    table_type="transient"
+                )
+                if not success:
+                    raise ValueError(
+                        f"Failed to write results to table {self._criteria_deletes_table_name}"
+                    )
+                logger.info(
+                    f"Wrote {nrows} rows and {nchunks} chunks to table {self._criteria_deletes_table_name}"
+                )
+        else:
+            logger.info("Results dataframe is empty, not applying")
+
 
 class ConnectResponse(SubscriptableBaseModel):
     """
@@ -1190,17 +1336,37 @@ class SnowflakeBillingEvent(BaseModel):
                 raise ValueError("start_timestamp must be timezone aware")
         return values
 
-class BillingEventRequest(BaseModel):
+class DailyBillingEventRequest(BaseModel):
     """
     Represents a request to provide billing events for that day.
+    These will occur at midnight, and cover the previous 24 hours.
+    Provides enough information for the plugin to create billing events for the day.
     """
 
-    billing_schedule: Literal["DAILY"]
-    has_active_inbound: bool
-    has_active_outbound: bool
+    billing_schedule: Literal["DAILY"] = "DAILY"
+    billable_connections_inbound: int = 0
+    billable_connections_outbound: int = 0
+    # below is deprecated
+    has_active_inbound: bool = False
+    has_active_outbound: bool = False
+
+class MonthlyBillingEventRequest(BaseModel):
+    """
+    Represents a request to provide billing events for that month.
+    These will occur at midnight on the first of each month, and cover the whole previous month.
+    Currently, these exist to provide a way to bill for the number of active ngrok connections
+    and associated data overages.
+    """
+
+    billing_schedule: Literal["MONTHLY"] = "MONTHLY"
+    distinct_active_ngrok_connections: int = 0
+    ngrok_data_usage_bytes: int = 0
+    # below is deprecated
+    has_active_inbound: bool = False
+    has_active_outbound: bool = False
 
 
-# BillingEventRequest = Annotated[Union[DailyBillingEventRequest,...],Field(discriminator='billing_schedule')]
+BillingEventRequest = Annotated[Union[DailyBillingEventRequest,MonthlyBillingEventRequest],Field(discriminator='billing_schedule')]
 
 
 class OmnataPlugin(ABC):
@@ -1218,6 +1384,12 @@ class OmnataPlugin(ABC):
         self._configuration_parameters: Optional[SyncConfigurationParameters] = None
         # the Snowpark session shouldn't need to be used, ordinarily
         self._session: Optional[Session] = None
+        self.disable_background_workers = False
+        """
+        disable_background_workers prevents the Python background workers from starting, which
+        take care of various status updates and loading enqueued results.
+        Only set this to True if you plan to do all of this yourself (e.g. in a Java stored proc)
+        """
     
 
     @abstractmethod
@@ -1388,6 +1560,39 @@ class OmnataPlugin(ABC):
         should be returned.
         """
         return []
+    
+    def omnata_standard_billing_events(self, request: BillingEventRequest,
+                                             initial_charge:Decimal,
+                                             additional_charge:Decimal) -> List[SnowflakeBillingEvent]:
+        """
+        Omnata's typical marketplace billing model (as at March 2024), is to bill the first daily event
+        per connection direction with the DAILY_ACTIVE_INITIAL class, and all subsequent connection
+        directions with the DAILY_ACTIVE_ADDITIONAL class.
+        """
+        sent_initial = False
+        events: List[SnowflakeBillingEvent] = []
+        for i in range(request.billable_connections_inbound + request.billable_connections_outbound):
+            if sent_initial is False:
+                events.append(
+                    SnowflakeBillingEvent(
+                        billing_class="DAILY_ACTIVE_INITIAL",
+                        billing_subclass="",
+                        timestamp=datetime.datetime.now(tz=datetime.timezone.utc),
+                        base_charge=initial_charge,
+                    )
+                )
+                sent_initial = True
+            else:
+                if additional_charge is not None and additional_charge > 0:
+                    events.append(
+                        SnowflakeBillingEvent(
+                            billing_class="DAILY_ACTIVE_ADDITIONAL",
+                            billing_subclass="",
+                            timestamp=datetime.datetime.now(tz=datetime.timezone.utc),
+                            base_charge=additional_charge,
+                        )
+                    )
+        return events
 
     def additional_loggers(self) -> List[str]:
         """
@@ -1681,9 +1886,16 @@ def __managed_inbound_processing_worker(
                 # logging this to the omnata_plugin logger in this way, 
                 # will cause it to automatically fail the appropriate stream
                 omnata_plugin_logger = getLogger("omnata_plugin")
-                omnata_plugin_logger.error(f"Error syncing stream {stream.stream_name}: {str(e)}", 
-                             exc_info=True, 
-                             extra={'stream_name':stream.stream_name})
+                try:
+                    omnata_plugin_logger.error(f"Error syncing stream {stream.stream_name}: {str(e)}", 
+                                exc_info=True, 
+                                extra={'stream_name':stream.stream_name})
+                except Exception as e2:
+                    # sometimes we get "Object of type MaxRetryError is not JSON serializable" or similar
+                    # so we need to handle that gracefully and just log it without the contents
+                    omnata_plugin_logger.error(f"{type(e).__name__} syncing stream {stream.stream_name}", 
+                                exc_info=True, 
+                                extra={'stream_name':stream.stream_name})
         except queue.Empty:
             logger.info("streams queue is empty")
             return
