@@ -30,6 +30,7 @@ from .omnata_plugin import (
     InboundSyncRequest,
     OmnataPlugin,
     OutboundSyncRequest,
+    DeadlineReachedException,
 )
 from .rate_limiting import ApiLimits, RateLimitState
 
@@ -139,21 +140,30 @@ class PluginEntrypoint:
                 rate_limit_state_all=rate_limit_state_all,
                 rate_limit_state_this_sync_and_branch=rate_limit_state_this_branch,
                 run_deadline=datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(minutes=request.time_limit_mins),
-                development_mode=False,
+                # for now, use the development mode flag to disable background workers
+                development_mode=self._plugin_instance.disable_background_workers
             )
             try:
                 self._plugin_instance._configuration_parameters = parameters
                 with HttpRateLimiting(outbound_sync_request, parameters):
                     self._plugin_instance.sync_outbound(parameters, outbound_sync_request)
+                if self._plugin_instance.disable_background_workers is False:
                     outbound_sync_request.apply_results_queue()
-                outbound_sync_request.apply_rate_limit_state()
+                    outbound_sync_request.apply_rate_limit_state()
+                if outbound_sync_request.deadline_reached:
+                    # if we actually hit the deadline, this is flagged by the cancellation checking worker and the cancellation
+                    # token is set. We throw it here as an error since that's currently how it flows back to the engine with a DELAYED state
+                    raise DeadlineReachedException()
             finally:
                 try:
                     # cancel the thread so we don't leave anything hanging around and cop a nasty error
                     outbound_sync_request._thread_cancellation_token.set()  # pylint: disable=protected-access
-                    outbound_sync_request._apply_results_task.join()  # pylint: disable=protected-access
-                    outbound_sync_request._cancel_checking_task.join()  # pylint: disable=protected-access
-                    outbound_sync_request._rate_limit_update_task.join()  # pylint: disable=protected-access
+                    if outbound_sync_request._apply_results_task is not None:
+                        outbound_sync_request._apply_results_task.join()  # pylint: disable=protected-access
+                    if outbound_sync_request._cancel_checking_task is not None:
+                        outbound_sync_request._cancel_checking_task.join()  # pylint: disable=protected-access
+                    if outbound_sync_request._rate_limit_update_task is not None:
+                        outbound_sync_request._rate_limit_update_task.join()  # pylint: disable=protected-access
                 except Exception as e:
                     logger.error(f"Error cleaning up threading: {str(e)}")
 
@@ -181,7 +191,8 @@ class PluginEntrypoint:
                 rate_limit_state_all=rate_limit_state_all,
                 rate_limit_state_this_sync_and_branch=rate_limit_state_this_branch,
                 run_deadline=datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(minutes=request.time_limit_mins),
-                development_mode=False,
+                # for now, use the development mode flag to disable background workers
+                development_mode=self._plugin_instance.disable_background_workers,
                 streams=list(request.streams_configuration.included_streams.values()),
                 omnata_log_handler=omnata_log_handler
             )
@@ -194,17 +205,25 @@ class PluginEntrypoint:
                 with HttpRateLimiting(inbound_sync_request, parameters):
                     self._plugin_instance.sync_inbound(parameters, inbound_sync_request)
                 logger.info("Finished invoking plugin")
-                inbound_sync_request.update_activity("Staging remaining records")
-                logger.info("Calling apply_results_queue")
-                inbound_sync_request.apply_results_queue()
-                inbound_sync_request.apply_rate_limit_state()
+                if self._plugin_instance.disable_background_workers is False:
+                    inbound_sync_request.update_activity("Staging remaining records")
+                    logger.info("Calling apply_results_queue")
+                    inbound_sync_request.apply_results_queue()
+                    inbound_sync_request.apply_rate_limit_state()
+                if inbound_sync_request.deadline_reached:
+                    # if we actually hit the deadline, this is flagged by the cancellation checking worker and the cancellation
+                    # token is set. We throw it here as an error since that's currently how it flows back to the engine with a DELAYED state
+                    raise DeadlineReachedException()
             finally:
                 # cancel the thread so we don't leave anything hanging around and cop a nasty error
                 try:
                     inbound_sync_request._thread_cancellation_token.set()  # pylint: disable=protected-access
-                    inbound_sync_request._apply_results_task.join()  # pylint: disable=protected-access
-                    inbound_sync_request._cancel_checking_task.join()  # pylint: disable=protected-access
-                    inbound_sync_request._rate_limit_update_task.join()  # pylint: disable=protected-access
+                    if inbound_sync_request._apply_results_task is not None:
+                        inbound_sync_request._apply_results_task.join()  # pylint: disable=protected-access
+                    if inbound_sync_request._cancel_checking_task is not None:
+                        inbound_sync_request._cancel_checking_task.join()  # pylint: disable=protected-access
+                    if inbound_sync_request._rate_limit_update_task is not None:
+                        inbound_sync_request._rate_limit_update_task.join()  # pylint: disable=protected-access
                 except Exception as e:
                     logger.error(f"Error cleaning up threading: {str(e)}")
             return_dict["errored_streams"] = list(
@@ -373,7 +392,7 @@ class PluginEntrypoint:
                 result = session.sql(event_query).collect()
                 logger.info(f"Billing event result: {result}")
             except Exception as e:
-                if '370001:3159209004' in str(e):
+                if '370001:3159209004' in str(e) or 'Application instance is not installed from listing' in str(e):
                     logger.warn('Billing event creation failed due to running internally to Omnata')
                 else:
                     raise e
@@ -401,7 +420,7 @@ class PluginEntrypoint:
                     }
             except Exception as exception:
                 logger.error(f"Error parsing secrets content for secret {other_secrets_name}: {str(exception)}")
-                raise ValueError("Error parsing secrets content:") from exception
+                raise ValueError(f"Error parsing secrets content: {str(exception)}") from exception
         return connection_secrets
     
     def ngrok_post_tunnel_fields(
